% HUANG-KECMAN-KOPRIVA BOOK
% CHAPTER 6
% MATLAB CODES FOR THE UNSUPERVISED LEARNING ALGORITHMS

%**************************************************************************
% 
% 1. Reads and displays two source images ('cameraman' and 'flower') shown in 
%    Figure 1. Corresponding Figure in the book is 6.1. 
% 2. Generates and displays mixed images shown in Figure 2. Corresponding Figure 
%    in the book is 6.2. 
% 3. Implements PCA and displays PCA transformed signals shown in Figures 3. 
%    Corresponding Figure in the book is 6.7.
% 4. Implements ICA and displays ICA transformed signal shown in Figures 4.
%    Corresponding Figure in the book is 6.26.
%
%**************************************************************************

clear all;	close all;	format compact

S1 = imread('cameraman','tif');             % reads source image s1
S2 = imread('flower','tif');                % reads source image s2

figure                                      % display source images (Fig. 6.1)
colormap('gray')
subplot(1,2,1)
imagesc(S1)
set(gca,'Xtick',[]);
set(gca,'Ytick',[]);
title('S1')
text(170,200,'Figure 6.1 Two source images')
axis('square')
subplot(1,2,2)
imagesc(S2)
set(gca,'Xtick',[]);
set(gca,'Ytick',[]);
title('S2')
axis('square')

[P Q] = size(S1);

s1 = double(reshape(S1,1,P*Q));                  % transform image into vector 
s2 = double(reshape(S2,1,P*Q));
s1 = s1 - mean(s1);
s2 = s2 - mean(s2);
S = [s1; s2];                            % matrix of source images

A = [2 1;1 1];                           % mixing matrix Eq.(6.2)

X=A*double(S);                           % matrix of mixed images
X1 = reshape(X(1,:),P,Q);                % transform vector into image
X2 = reshape(X(2,:),P,Q);

figure                                   % display mixed images  (Fig. 6.2)
colormap('gray')
subplot(1,2,1)
imagesc(X1)
title('X1')
set(gca,'Xtick',[]);
set(gca,'Ytick',[]);
axis('square')
text(175,200,'Figure 6.2 Two mixed images')
subplot(1,2,2)
imagesc(X2)
set(gca,'Xtick',[]);
set(gca,'Ytick',[]);
title('X2')
axis('square')

% PCA transform
Rx = cov(X');                         % data covariance matrix
[E,LAM] = eig(Rx);                    % eigen-decomposition of Rx
lam = diag(LAM);                      % extract eigen-values
[lam,ind] = sort(lam);
E = E(:,ind);
lam = sqrt(1./lam);                   % square root of inverse
LAM = diag(lam);                      % back into diagonal matrix form
Z = LAM*E'*X;                           % PCA transform

Z1 = reshape(Z(1,:),P,Q);            % transform vector into image
Z2 = reshape(Z(2,:),P,Q);            % transform vector into image

figure                               % display PCA recovered images  (Fig. 6.7)
colormap('gray')
subplot(1,2,1)
imagesc(-Z1)                         % correct for sign indeterminacy
set(gca,'Xtick',[]);
set(gca,'Ytick',[]);
title('Z1')
axis('square')
text(50,200,'Figure 6.7 Source images reconstructed by the PCA transform')

subplot(1,2,2)
imagesc(-Z2)                         % correct for sign indeterminacy
set(gca,'Xtick',[]);
set(gca,'Ytick',[]);
title('Z2')                
axis('square')

% ICA transform by the minimum of the mutual information
gain = 4e-8;                        % learning gain
theta = 3;                          % Gaussian exponent for sub-Gaussian signals
ITMAX = 700;                        % maximal number of iterations
it = 0;                             % iteration counter
I = eye(2);                         % identity matrix
W = I;                              % initial value for de-mixing matrix
h = waitbar(0,'ICA transform. Please wait...');
while (it < ITMAX)                  % iterate until it > ITMAX
  it = it +1;
  Y = W*X;                          % ICA transform Eq. (6.3)
  for i = 1:2
      fi(i,:) = sign(Y(i,:)).*power(abs(Y(i,:)),theta-1); % score function Eq.(6.48)
  end
  F = I - fi*Y'/(P*Q);                 % Eq. (6.42)
  W = W + gain*F*W;                    % Eq. (6.42)
   waitbar(it/ITMAX,h)
end
close (h)

Y1 = reshape(Y(1,:),P,Q);           % transform vector into image
Y2 = reshape(Y(2,:),P,Q);           % transform vector into image

figure                              % display ICA recovered images  (Fig. 6.26)
colormap('gray')
subplot(1,2,1)
imagesc(Y1)                        
set(gca,'Xtick',[]);
set(gca,'Ytick',[]);
title('Y1')
axis('square')
text(20,200,'Figure 6.26 Source images reconstructed by the ICA algorithm (6.42)')

subplot(1,2,2)
imagesc(Y2)                        
set(gca,'Xtick',[]);
set(gca,'Ytick',[]);
title('Y2')                
axis('square')
pause

figure                             % histograms of the source images (Fig. 6.8)
colormap('gray')
subplot(1,2,1)
hist(double(s1))      
grid
title('s1')
text(-115,-1150,'Figure 6.8 Histograms of two source images shown in Fig. 6.1')
subplot(1,2,2)
hist(double(s2))
grid
title('s2')             
colormap([0.5 0.5 0.5])

figure                            % histograms of the mixed images (Fig. 6.9)
colormap('gray')
subplot(1,2,1)
hist(X(1,:))      
grid
title('X1')
text(-275,-800,'Figure 6.9 Histograms of two mixed images shown in Fig. 6.2')
subplot(1,2,2)
hist(X(2,:))
grid
title('X2')             
colormap([0.5 0.5 0.5])

figure                            % histograms of the PCA de-mixed images (Fig. 6.10)
colormap('gray')
subplot(1,2,1)
hist(-Z(1,:))      
grid
title('Z1')
text(-2.5,-1100,'Figure 6.10 Histograms of two PCA images shown in Fig. 6.7')
subplot(1,2,2)
hist(-Z(2,:))
grid
title('Z2')             
colormap([0.5 0.5 0.5])

figure                            % histograms of the ICA de-mixed images (Fig. 6.29)
colormap('gray')
subplot(1,2,1)
hist(Y(1,:))      
grid
title('Y1')
text(-55,-1150,'Figure 6.29 Histograms of the images reconstructed by the ICA algorithm (6.42)')
subplot(1,2,2)
hist(Y(2,:))
grid
title('Y2')             
colormap([0.5 0.5 0.5])
