% HUANG-KECMAN-KOPRIVA BOOK
% CHAPTER 6
% MATLAB CODES FOR THE UNSUPERVISED LEARNING ALGORITHMS

%**************************************************************************
% 
% 1. Reads and displays two speech source signals shown in Figure 1. 
%    Corresponding Figure in the book is 6.3.
% 2. Generates and displays mixed speech signals shown in Figures 2.
%    Corresponding Figure in the book is 6.4.
% 3. Implements PCA tarnsform and dispalys PCA recovered signals shown in Figure 3.
%    Corresponding Figure in the book is 6.12.
% 4. Implements ICA transform and displays ICA recovered signals shown in
%    Figure 4. Corresponding Figure in the book is 6.30.
% 5. Estimates and displays histograms of the source signals, mixed signals, PCA 
%    recovered signals and ICA recovered signals shown in Figures 5, 6, 7 and 8. 
%    Corresponding Figure in the book are 6.13, 6.14, 6.15 and 6.31 respectively.
%
%**************************************************************************

clear all;	close all;	format compact

s1(1,:) = wavread('speech_1')';         % reads source speech signal s1
s2(1,:) = wavread('speech_2')';         % reads source speech signal s2
Fs = 16000;                             % sampling frequency 16 kHz
Ts = 1/Fs;
T = length(s1);                         % length of the data record
t = 0:Ts:(T-1)*Ts;                       % generate independent time variable

figure(1)                                  % displays source speech signals (Fig. 6.3)
subplot(1,2,1)
plot(t,s1,'k')
grid
text(3.05,-0.45,'time [s]')
title('s1')
axis([0 4 -0.5 1])
text(2.25,-0.625,'Figure 6.3 Two source speech signals')
subplot(1,2,2)
plot(t,s2,'k')
grid
text(3.05,-0.45,'time [s]')
title('s2')
axis([0 4 -0.5 1])

S = [s1; s2];                         % matrix of source signals

A = [2 1;1 1];                        % mixing matrix Eq.(6.2)

X=A*S;                               % matrix of mixed speech signals

figure(2)                              % displays mixed speech signals (Fig. 6.4)
subplot(1,2,1)
plot(t,X(1,:),'k')
grid
text(3.05,-0.9,'time [s]')
title('x1')
axis([0 4 -1 2])
text(2.5,-1.25,'Figure 6.4 Two mixed speech signals')
subplot(1,2,2)
plot(t,X(2,:),'k')
grid
text(3.05,-0.9,'time [s]')
title('x2')
axis([0 4 -1 2])

% whitening transform
Rx = cov(X');                         % data covariance matrix
[E,LAM] = eig(Rx);                    % eigen-decomposition of Rx
lam = diag(LAM);                      % extract eigen-values
[lam,ind] = sort(lam);
E = E(:,ind);
lam = sqrt(1./lam);                   % square root of inverse
LAM = diag(lam);                      % back into diagonal matrix form
Z = LAM*E'*X;                         % the whitening transform

figure(3)                                % displays PCA recovered speech signals (Fig. 6.12)
subplot(1,2,1)  
plot(t,-Z(1,:),'k')
grid
text(3.05,-3.75,'time [s]')
title('z1')
axis([0 4 -4 8])
text(-0.15,-5,'Figure 6.12 Source speech signals reconstructed by whitening (PCA) transform')
subplot(1,2,2)
plot(t,-Z(2,:),'k')
grid
text(3.05,-3.75,'time [s]')
title('z2')
axis([0 4 -4 8])

gain = 0.4;                         % learning gain
theta = 1.5;                        % Gaussian exponent for sub-Gaussian signals
ITMAX = 400;                        % maximal number of iterations
it = 0;                             % iteration counter
I = eye(2);                         % identity matrix
W = I;                              % initial value for de-mixing matrix
h = waitbar(0,'ICA transform. Please wait...');
while (it < ITMAX)                  % iterate until it > ITMAX
  it = it +1;
  Y = W*Z;                          % ICA transform Eq. (6.3)
  for i = 1:2
      fi(i,:) = sign(Y(i,:)).*power(abs(Y(i,:)),theta-1); % score function Eq.(6.48)
  end
  F = I - fi*Y'/T;                     % Eq. (6.42)
  W = W + gain*F*W;                    % Eq. (6.42)
  waitbar(it/ITMAX,h)
end
close (h)

figure(4)                              % display ICA recovered speech signals  (Fig. 6.30)
subplot(1,2,1)  
plot(t,-Y(1,:),'k')
grid
text(3.05,-3.75,'time [s]')
title('y1')
axis([0 4 -4 8])
text(-0.1,-5,'Figure 6.30 Source speech signals reconstructed by the ICA algorithm (6.42)')
subplot(1,2,2)
plot(t,-Y(2,:),'k')
grid
text(3.05,-3.75,'time [s]')
title('y2')
axis([0 4 -8 12])
         

figure(5)                               % histograms of the source signals (Fig. 6.13)
subplot(1,2,1)
hist(s1)      
grid
title('s1')
text(-0.25,-2600,'Figure 6.13 Histograms of two speech signals shown in Fig. 6.3')
subplot(1,2,2)
hist(s2)
grid
title('s2')             
colormap([0.5 0.5 0.5])

figure(6)                               % histograms of the mixed signals (Fig. 6.14)
subplot(1,2,1)
hist(X(1,:))      
grid
title('X1')
text(-0.75,-2600,'Figure 6.14 Histograms of two mixed speech signals shown in Fig. 6.4')
subplot(1,2,2)
hist(X(2,:))
grid
title('X2')             
colormap([0.5 0.5 0.5])

figure(7)                            % histograms of the PCA de-mixed signals (Fig. 6.15)
subplot(1,2,1)
hist(Z(1,:))      
grid
title('Z1')
text(-11,-3000,'Figure 6.15 Histograms of two PCA recovered speech signals shown in Fig. 6.12')
subplot(1,2,2)
hist(Z(2,:))
grid
title('Z2')             
colormap([0.5 0.5 0.5])

figure(8)                            % histograms of the ICA de-mixed signals (Fig. 6.31)
subplot(1,2,1)
hist(-Y(1,:))      
grid
title('Y1')
text(-7,-2600,'Figure 6.31 Histograms of the speech signals recovered by the ICA algorithm (6.42)')
subplot(1,2,2)
hist(-Y(2,:))
grid
title('Y2')             
colormap([0.5 0.5 0.5])
